
import jsonfile from "jsonfile";
import cheerio from "cheerio";
import fs from "fs";
import { getDomain } from "tldts";
import { Browser, launch }  from "puppeteer";
import axios from "axios";
import { Dobey, LDJsonParser } from "./modules/websites";

interface IProduct {
    name: string;
    blogUrl: string;
    url: string;
    domain: string;
    inStock: boolean;
};

// Globals
let browser: Browser = null;
let sitemapUrl: string = "";
let allProducts: Array<IProduct> = [];
let allBlogUrls: Array<string> = [];

// debugStart();
start();

async function debugStart() {
    browser = await launch({ headless: false });
    console.log(await crawlProductStock("https://www.kabels.nl/nl_nl/ewent-ew3240-draadloze-multi-connect-muis-600-2400-dpi-zwart-34863350.html"));
}

async function start() {
    try {
        console.log(`---- ContentEggStockCrawler ----`);
        const urls: Array<string> = [];
        const sitemapUrls: Array<string> = [];

        // Get sitemap url from process params
        let params = process.argv.slice(2);
        sitemapUrl = params[0];

        // Crawl Sitemap
        console.log(`Crawling input sitemap: ${sitemapUrl}`);
        const sitemapRequest = await axios.get(sitemapUrl);
        const $ = cheerio.load(sitemapRequest.data, { xmlMode: true });

        $('loc').each(function() {
            const url = $(this).text();

            if (url.endsWith(".xml")) {
                if (!sitemapUrls.includes(url)) {
                    console.log("Found sitemap url: " + url);
                    sitemapUrls.push(url);
                }   
            } else {
                if (!urls.includes(url)) {
                    urls.push(url);
                }
            }
        });


        for (const sitemapUrl of sitemapUrls) {
            console.log(`Crawling crawled sitemap: ${sitemapUrl}`);
            const sitemapRequest = await axios.get(sitemapUrl);
            const $ = cheerio.load(sitemapRequest.data, { xmlMode: true });

            $('loc').each(function() {
                const url = $(this).text();

                if (!urls.includes(url)) {
                    urls.push(url);
                }
            });
        }

        console.log(`Crawled ${urls.length} urls from sitemaps.`);
        
        console.log(`Starting crawler browser..`);
        browser = await launch({ headless: true });

        console.log(`Initialization done.`);
        console.log(`------------------------------------- \n`);

        for (let url of urls) {
            try {
                allBlogUrls.push(url);
                await crawlUrl(url); 
            } catch (error) {
                continue;   
            }
        }

        await browser.close();
        await generateReport();
    } catch (error) {
        console.error(`An Error Occured!`, error);    
    }
}

async function crawlUrl(url: string) {
    try {
        console.log(`Crawling url: ${url}`);
    
        // Open new page and goto url
        const page = await browser.newPage();
        await page.goto(url, { waitUntil: 'networkidle2' });

        // Get html from page
        const html = await page.evaluate(() => document.body.innerHTML);

        // Close page
        await page.close();

        // Load html in cheerio object
        const $ = cheerio.load(html);

        let products: Array<any> = [];

        if ($(".row-products").length >= 1) {
            console.log(`- Detected ${$(".row-products").length} content egg row type products.`);

            $(".row-products .cegg-list-logo-title a").each((index, element) => {
                let productUrl = $(element).attr("href");
                let name = $(element).html().trim();
                
                products.push({
                    name: name,
                    blogUrl: url,
                    url: productUrl,
                    domain: "",
                    inStock: false,
                });
            });
        }
        
        if ($("div.egg-container.egg-item").length >= 1) {
            console.log(`- Detected ${$("div.egg-container.egg-item").length} content egg large type products.`);

            $("div.egg-container.egg-item").each((index, element) => {
                let productUrl = $(element).find(".cegg-btn-row a").first().attr("href");
                let name = $(element).find("h2").first().html().trim();
                
                products.push({
                    name: name,
                    blogUrl: url,
                    url: productUrl,
                    domain: "",
                    inStock: false,
                });
            });
        }

        console.log("- Checking product stocks...");

        for (let index in products) {
            try {
                let status = await crawlProductStock(products[index].url);

                products[index].domain = status[0]; 
                products[index].url = status[1];
                products[index].inStock = status[2];
            
                if (products[index].inStock) {
                    console.log(`   [IN STOCK] ${products[index].name} - ${products[index].domain}`);
                } else {
                    console.log(`   [OUT OF STOCK] ${products[index].name} - ${products[index].domain} - ${products[index].url}`);
                }

                allProducts.push(products[index]);
            } catch (error) {
                console.log("- Skipping product...");
                continue;
            }
        }

        console.log("");
        console.log("");

    } catch (error) {
        console.log("-- Error while trying to crawl page! Skipping...");   
    }
}

async function crawlProductStock(url: string) {
    try {
        // Open new page and goto url
        const page = await browser.newPage();
        await page.goto(url, { waitUntil: 'networkidle2' });

        // Get html from page
        const html = await page.evaluate(() => document.body.innerHTML);

        // Load html in cheerio object
        const $ = cheerio.load(html);

        // Get domain from url
        const domain = getDomain(page.url());

        // Close page
        await page.close();

        switch (domain) {
            case "bol.com":
                return [domain, page.url(), await LDJsonParser.check(html)];

            case "petsplace.nl":
                return [domain, page.url(), await LDJsonParser.check(html)];

            case "dobey.nl":
                return [domain, page.url(), await Dobey.check(html)];

            case "brekz.nl":
                return [domain, page.url(), await LDJsonParser.check(html)];

            case "hondenbed.nl":
                return [domain, page.url(), await LDJsonParser.check(html)];
            
            case "petsonline.nl":
                return [domain, page.url(), await LDJsonParser.check(html)];
            
            default:
                console.error(`-- ${domain} is not an supported website! Cannot check stock!`);
                return [domain, page.url(), false];
        }
    } catch (error) {
        console.error(error);
        console.log("-- Error while trying to crawl page! Skipping...");
    }
}

async function generateReport() {
    console.log("Generating report...");

    let today: Date = new Date();
    let datecode: string = today.getFullYear().toString() + "-" + (today.getMonth() + 1).toString() + "-" + today.getDay().toString();
    let domain: string = getDomain(sitemapUrl);
    let file = `${domain}_${datecode}_report.txt`;

    let totalProducts = 0;
    let totalOutStock = 0;
    for (let product of allProducts) {
        totalProducts++;

        if (product.inStock == false) {
            totalOutStock++;
        }
    }

    appendLn(file, `Content Egg Product Stock Crawler Report\n`);
    appendLn(file, `Total Blog Urls: ${allBlogUrls.length}\n`);
    appendLn(file, `Total Products: ${totalProducts}\n`);
    appendLn(file, `Total Out of Stock: ${totalOutStock}\n`);
    appendLn(file, ``);

    let lastBlogUrl = "";
    for (let product of allProducts) {
        if (product.blogUrl != lastBlogUrl) {
            appendLn(file, `\n\n`);
            appendLn(file, `${product.blogUrl}\n`);
            appendLn(file, `----------------------------------\n`);
        }

        if (product.inStock == false) {
            appendLn(file, `${product.domain} - ${product.name} - ${product.url}\n`);
        }

        lastBlogUrl = product.blogUrl;
    }

    console.log("Report generated.");
}

function appendLn(file: string, text: string): Promise<void> {
    return new Promise((resolve, reject) => {
        fs.appendFile(file, text + "\n", (err) => {
            if (err) {
                reject(err);
            } else {
                resolve();
            }
        });
    });
}